class MdXlsStyleRegistry(object):

    default_formats = {
        'double_emphasis': {'bold': True},
        'emphasis': {'italic': True},
        'strikethrough': {'font_strikeout': True},
        'codespan': {'font_name': 'Courier'},
        'h1': {'font_size': 30},
        'h2': {'font_size': 25},
        'h3': {'font_size': 20},
        'h4': {'font_size': 15},
        'h5': {'font_size': 14},
        'h6': {'font_size': 13},
    }

    def __init__(self, workbook):
        self.workbook = workbook
        self.stylereg = {}

    def use_style(self, mdnames):

        if not isinstance(mdnames, list):
            mdnames = [mdnames]

        mdname = '-'.join(mdnames)

        if not mdname in self.stylereg:

            style = self._create_style(mdnames)

            self.stylereg[mdname] = style

        return self.stylereg[mdname]

    def _create_style(self, mdnames):

        d = {}

        for submdname in mdnames:

            if submdname in self.default_formats:
                d = {**d, **self.default_formats[submdname]}

        if len(d) > 0:
            return self.workbook.add_format(d)

        return ''



import copy
from io import BytesIO
import re
import base64

from nbconvert.exporters import Exporter

from bs4 import BeautifulSoup
from bs4.element import NavigableString, Tag
import xlsxwriter

import mistune
from .mdrenderer import Md2XLSRenderer, \
    MdStyleInstructionCell, MdStyleInstructionText, MdStyleInstructionLink, MdStyleInstructionList
from .mdxlsstyles import MdXlsStyleRegistry

try:
    import cv2 # Prefer Open CV2 but don't put in requirements.txt because it can be difficult to install
    import numpy as np
    usecv2 = True
except ImportError:
    import png # PyPNG is a pure-Python PNG manipulator
    usecv2 = False


class XLSExporter(Exporter):
    """
    XLSX custom exporter
    """

    # If this custom exporter should add an entry to the
    # "File -> Download as" menu in the notebook, give it a name here in the
    # `export_from_notebook` class member
    export_from_notebook = "Excel Spreadsheet"

    def __init__(self, config=None, **kw):
        """
        Public constructor

        Parameters
        ----------
        config : :class:`~traitlets.config.Config`
            User configuration instance.
        `**kw`
            Additional keyword arguments passed to parent __init__

        """

        super(XLSExporter, self).__init__(config=config, **kw)

        self.msxlsstylereg = None
        self.workbook = None
        self.row = 0

    def _file_extension_default(self):
        """
        The new file extension is `.xlsx`
        """
        return '.xlsx'

    def from_notebook_node(self, nb, resources=None, **kw):
        """
        Convert a notebook from a notebook node instance.
        Parameters
        ----------
        nb : :class:`~nbformat.NotebookNode`
          Notebook node (dict-like with attr-access)
        resources : dict
          Additional resources that can be accessed read/write by
          preprocessors and filters.
        `**kw`
          Ignored
        """
        nb_copy = copy.deepcopy(nb)
        resources = self._init_resources(resources)

        if 'language' in nb['metadata']:
            resources['language'] = nb['metadata']['language'].lower()

        # Preprocess
        nb_copy, resources = super(XLSExporter, self).from_notebook_node(nb, resources, **kw)

        output = BytesIO()
        self.workbook = xlsxwriter.Workbook(output)

        self.msxlsstylereg = MdXlsStyleRegistry(self.workbook)

        self.worksheet = self.workbook.add_worksheet()

        self.row = 0
        for cellno, cell in enumerate(nb_copy.cells):
            self.worksheet.write(self.row, 0, str(cellno+1))

            if cell.cell_type == 'markdown':
                self._write_markdown(cell.source)

            elif cell.cell_type == 'code':
                for o in cell.outputs:
                    self._write_code(o)

            self.row += 1

        self.workbook.close()

        xlsx_data = output.getvalue()

        return xlsx_data, resources

    def _write_code(self, o):
        """
        Main handler for code cells
        :param o:
        :return:
        """
        if o.output_type == 'stream':
            self._write_textplain(o.text)
        if o.output_type in ('display_data', 'execute_result'):
            if 'text/html' in o.data:
                self._write_texthtml(o.data['text/html'])
            elif 'image/png' in o.data:
                width, height = 0, 0
                if 'image/png' in o.metadata and set(o.metadata['image/png'].keys()) == {'width', 'height'} :
                    width, height = o.metadata['image/png']['width'], o.metadata['image/png']['height']
                self._write_image(o.data['image/png'], width, height)
            elif 'application/json' in o.data:
                self._write_textplain(o.data['application/json'])
            elif 'text/plain' in o.data:
                self._write_textplain(o.data['text/plain'])

    ###
    # Sub-handlers for code cells

    def _write_textplain(self, text):
        lines = text.split("\n")
        for l in lines:
            self.worksheet.write(self.row, 1, l)
            self.row += 1

    # HTML functions start here

    def _write_texthtml(self, html):
        soup = BeautifulSoup(html, 'html.parser')
        self._write_soup(soup)

    def _write_soup(self, soup):
        s = ''
        for child in soup.children:

            if isinstance(child, NavigableString):
                s += child.string

            elif isinstance(child, Tag):

                if len(s) > 0:
                    # Write accumulated string first
                    re.sub('\s+', ' ', s)
                    s = s.strip()
                    if len(s) > 0:
                        self.worksheet.write(self.row, 1, s.strip())
                        self.row += 1
                        s = ''

                if child.name in ('div', 'body', 'span', 'p'):
                    self._write_soup(child)

                elif child.name == 'table':
                    self._write_htmltable(soup)

    def _write_htmltable(self, soup):
        for tablerow in soup('tr'):
            col = 1
            for child in tablerow.children:
                if isinstance(child, Tag):
                    if child.name == 'th' or child.name == 'td':
                        s = child.get_text()
                        try:
                            f = float(s)
                            self.worksheet.write_number(self.row, col, f)
                        except ValueError:
                            self.worksheet.write(self.row, col, s)

                        col += 1
            self.row += 1

    # Image handler

    def _write_image(self, image, want_width, want_height):

        image = base64.b64decode(image)

        image_data = BytesIO(image)

        if usecv2:
            nparr = np.fromstring(image, np.uint8)
            img = cv2.imdecode(nparr, cv2.IMREAD_ANYCOLOR)
            height, width = img.shape[:2]
        else:
            img = png.Reader(image_data)
            (width, height, _, _) = img.asDirect()

        x_scale, y_scale = 1.0, 1.0

        if want_height > 0 and height > 0:
            y_scale = want_height / height

        if want_width > 0 and width > 0:
            x_scale = want_width / width

        self.row += 1

        self.worksheet.insert_image(self.row, 1, 'image.png',
                                    {'image_data': image_data, 'x_scale': x_scale, 'y_scale': y_scale})
        self.worksheet.set_row(self.row, height*y_scale)

        self.row += 1

    # Markdown handler

    def _write_markdown(self, md):
        try:
            markdown = mistune.Markdown(renderer=Md2XLSRenderer())
            lines = markdown(md)

            all_o = []

            for l in lines:
                in_softnewline = False
                is_indented = False
                link_url = ''
                already_outputted_text = False
                cell_format_mdname = ''
                o = []
                mdtextstylenames = []
                for i,s in enumerate(l):
                    if isinstance(s, MdStyleInstructionText):
                        mdtextstylenames += [s.mdname]

                    elif isinstance(s, MdStyleInstructionCell):
                        cell_format_mdname = s.mdname

                    elif isinstance(s, MdStyleInstructionLink):
                        if already_outputted_text:
                            all_o.append([o, cell_format_mdname, link_url, is_indented])
                            o = []
                            already_outputted_text = False
                        in_softnewline = True
                        link_url = s.link

                    elif isinstance(s, MdStyleInstructionList):
                        if already_outputted_text:
                            all_o.append([o, cell_format_mdname, link_url, is_indented])
                            o = []
                            already_outputted_text = False
                        in_softnewline = True
                        link_url = ''
                        is_indented = True

                    elif len(s) > 0:
                        if len(mdtextstylenames) > 0 or (cell_format_mdname != '' and i >= 2):
                            format = self.msxlsstylereg.use_style([cell_format_mdname] + mdtextstylenames)
                            o.append(format)
                            mdtextstylenames = []

                        o.append(s)
                        already_outputted_text = True

                        if in_softnewline:
                            all_o.append([o, cell_format_mdname, link_url, is_indented])
                            o = []
                            already_outputted_text = False
                            in_softnewline = False
                            is_indented = False
                            link_url = ''

                if len(o) > 0:
                    all_o.append([o, cell_format_mdname, link_url, is_indented])

            for o, cell_format_mdname, link_url, is_indented in all_o:

                is_indented = int(is_indented)

                if cell_format_mdname != '':
                    o.append(self.msxlsstylereg.use_style(cell_format_mdname))

                if link_url != '':
                    if len(o) >= 2:
                        self.worksheet.write_url(self.row, 1+is_indented, link_url, o[1], o[0])
                    elif len(o) == 1:
                        self.worksheet.write_url(self.row, 1+is_indented, link_url, None, o[0])

                else:

                    if len(o) > 2:
                        self.worksheet.write_rich_string(self.row, 1+is_indented, *o)
                    elif len(o) == 2:
                        if isinstance(o[0], xlsxwriter.format.Format) and not isinstance(o[1], xlsxwriter.format.Format):
                            self.worksheet.write(self.row, 1+is_indented, o[1], o[0])
                        elif not isinstance(o[0], xlsxwriter.format.Format) and not isinstance(o[1], xlsxwriter.format.Format):
                            self.worksheet.write(self.row, 1+is_indented, o[0] + ' ' + o[1])
                        else:
                            self.worksheet.write(self.row, 1+is_indented, o[0], o[1])
                    elif len(o) == 1 and not isinstance(o[0], xlsxwriter.format.Format):
                        self.worksheet.write(self.row, 1+is_indented, o[0])

                self.row += 1

        except Exception as e:
            print('Markdown Exception: ', e)
            self._write_textplain(md)



import re

from mistune import Renderer, escape, escape_link


class MdStyleInstruction(object):

    softnewline = False

    def __init__(self, mdname):
        self.mdname = mdname

    def __repr__(self):
        params = ['{}={}'.format(a, getattr(self,a)) for a in dir(self) if not a.startswith('__') and not callable(getattr(self,a))]
        return '<{} {}>'.format(type(self), ", ".join(params))


class MdStyleInstructionCell(MdStyleInstruction):
    pass


class MdStyleInstructionText(MdStyleInstruction):
    pass


class MdStyleInstructionLink(MdStyleInstruction):

    softnewline = True

    def __init__(self, link):
        super(MdStyleInstructionLink, self).__init__('link')
        self.link = link


class MdStyleInstructionList(MdStyleInstruction):

    softnewline = True

    def __init__(self):
        super(MdStyleInstructionList, self).__init__('list_item')


class Md2XLSRenderer(Renderer):

    def placeholder(self):
        """Returns the default, empty output value for the renderer.
        All renderer methods use the '+=' operator to append to this value.
        Default is a string so rendering HTML can build up a result string with
        the rendered Markdown.
        Can be overridden by Renderer subclasses to be types like an empty
        list, allowing the renderer to create a tree-like structure to
        represent the document (which can then be reprocessed later into a
        separate format like docx or pdf).
        """
        return []

    ### Block-level functions

    def block_code(self, code, lang=None):
        """Rendering block level code. ``pre > code``.
        :param code: text content of the code block.
        :param lang: language of the given code.
        """
        code = code.rstrip('\n')
        return ["<code>"] + code

    def block_quote(self, text):
        """Rendering <blockquote> with the given text.
        :param text: text content of the blockquote.
        """
        return ["<blockquote>"] + text

    def block_html(self, html):
        """Rendering block level pure html content.
        :param html: text content of the html snippet.
        """
        if self.options.get('skip_style') and \
                html.lower().startswith('<style'):
            return ['']
        if self.options.get('escape'):
            return [escape(html)]
        return [html]

    def header(self, text, level, raw=None):
        """Rendering header/heading tags like ``<h1>`` ``<h2>``.
        :param text: rendered text content for the header.
        :param level: a number for the header level, for example: 1.
        :param raw: raw text content of the header.
        """
        return [[MdStyleInstructionCell('h{}'.format(level))] + text]

    def hrule(self):
        """Rendering method for ``<hr>`` tag."""
        return [MdStyleInstructionCell('hrule')]

    def list(self, body, ordered=True):
        """Rendering list tags like ``<ul>`` and ``<ol>``.
        :param body: body contents of the list.
        :param ordered: whether this list is ordered or not.
        """
        return body

    def list_item(self, text):
        """Rendering list item snippet. Like ``<li>``."""
        return [[MdStyleInstructionList(), *text]]

    def paragraph(self, text):
        """Rendering paragraph tags. Like ``<p>``."""
        return [text]

    def table(self, header, body):
        """Rendering table element. Wrap header and body in it.
        :param header: header part of the table.
        :param body: body part of the table.
        """
        return header + body

    def table_row(self, content):
        """Rendering a table row. Like ``<tr>``.
        :param content: content of current table row.
        """
        return ['<tr>\n%s</tr>\n'] + content

    def table_cell(self, content, **flags):
        """Rendering a table cell. Like ``<th>`` ``<td>``.
        :param content: content of current table cell.
        :param header: whether this is header or not.
        :param align: align of current table cell.
        """
        return content


    ### Span-level functions

    def double_emphasis(self, text):
        """Rendering **strong** text.
        :param text: text content for emphasis.
        """
        return [MdStyleInstructionText('double_emphasis')] + text

    def emphasis(self, text):
        """Rendering *emphasis* text.
        :param text: text content for emphasis.
        """
        return [MdStyleInstructionText('emphasis')] + text

    def codespan(self, text):
        """Rendering inline `code` text.
        :param text: text content for inline code.
        """
        text = escape(text.rstrip(), smart_amp=False)
        return [MdStyleInstructionText('codespan')] + text

    def linebreak(self):
        """Rendering line break like ``<br>``."""

        return ["<lb>"]

    def strikethrough(self, text):
        """Rendering ~~strikethrough~~ text.
        :param text: text content for strikethrough.
        """
        return [MdStyleInstructionText('strikethrough')] + text

    def text(self, text):
        """Rendering unformatted text.
        :param text: text content.
        """
        return [escape(text)]

    def escape(self, text):
        """Rendering escape sequence.
        :param text: text content.
        """
        return [escape(text)]

    def autolink(self, link, is_email=False):
        """Rendering a given link or email address.
        :param link: link content or email address.
        :param is_email: whether this is an email or not.
        """
        text = link = escape_link(link)
        if is_email:
            link = 'mailto:%s' % link
        return [MdStyleInstructionLink(link)] + text

    def link(self, link, title, text):
        """Rendering a given link with content and title.
        :param link: href link for ``<a>`` tag.
        :param title: title content for `title` attribute.
        :param text: text content for description.
        """
        link = escape_link(link)
        return [MdStyleInstructionLink(link)] + text

    def image(self, src, title, text):
        """Rendering a image with title and text.
        :param src: source link of the image.
        :param title: title text of the image.
        :param text: alt text of the image.
        """
        src = escape_link(src)
        text = escape(text, quote=True)
        if title:
            title = escape(title, quote=True)
            html = '<img src="%s" alt="%s" title="%s"' % (src, text, title)
        else:
            html = '<img src="%s" alt="%s"' % (src, text)
        if self.options.get('use_xhtml'):
            return '%s />' % html
        return '%s>' % html

    def inline_html(self, html):
        """Rendering span level pure html content.
        :param html: text content of the html snippet.
        """
        if self.options.get('escape'):
            return [escape(html)]
        return [html]

    def newline(self):
        """Rendering newline element."""
        return ['<new line>']

    def footnote_ref(self, key, index):
        """Rendering the ref anchor of a footnote.
        :param key: identity key for the footnote.
        :param index: the index count of current footnote.
        """
        html = (
                   '<sup class="footnote-ref" id="fnref-%s">'
                   '<a href="#fn-%s">%d</a></sup>'
               ) % (escape(key), escape(key), index)
        return html

    def footnote_item(self, key, text):
        """Rendering a footnote item.
        :param key: identity key for the footnote.
        :param text: text content of the footnote.
        """
        back = (
                   '<a href="#fnref-%s" class="footnote">&#8617;</a>'
               ) % escape(key)
        text = text.rstrip()
        if text.endswith('</p>'):
            text = re.sub(r'<\/p>$', r'%s</p>' % back, text)
        else:
            text = '%s<p>%s</p>' % (text, back)
        html = '<li id="fn-%s">%s</li>\n' % (escape(key), text)
        return html

    def footnotes(self, text):
        """Wrapper for all footnotes.
        :param text: contents of all footnotes.
        """
        html = '<div class="footnotes">\n%s<ol>%s</ol>\n</div>\n'
        return html % (self.hrule(), text)




from distutils.util import convert_path
from setuptools import setup, find_packages

module = 'nb2xls'

# get version from __meta__
meta_ns = {}
path = convert_path(module+'/__meta__.py')
with open(path) as meta_file:
    exec(meta_file.read(), meta_ns)

# read requirements.txt
with open('requirements.txt', 'r') as f:
    content = f.read()
li_req = content.split('\n')
install_requires = [e.strip() for e in li_req if len(e)]


name = module
name_url = name.replace('_', '-')

packages = [module]
version = meta_ns['__version__']
description = 'Export Jupyter notebook as an Excel xls file.'
long_description = 'Export Jupyter notebook as an Excel xls file.'
author = 'ideonate'
author_email = 'dan@ideonate.com'
# github template
url = 'https://github.com/{}/{}'.format(author,
                                        name_url)
download_url = 'https://github.com/{}/{}/tarball/{}'.format(author,
                                                            name_url,
                                                            version)
keywords = ['jupyter',
            'nbconvert',
            ]
license = 'MIT'
classifiers = ['Development Status :: 4 - Beta',
               'License :: OSI Approved :: MIT License',
               'Programming Language :: Python :: 3.5',
               'Programming Language :: Python :: 3.6',
               'Programming Language :: Python :: 3.7'
               ]
include_package_data = True

#install_requires = install_requires
zip_safe = False


# ref https://packaging.python.org/tutorials/distributing-packages/
setup(
    name=name,
    version=version,
    packages=packages,
    author=author,
    author_email=author_email,
    description=description,
    long_description=long_description,
    url=url,
    download_url=download_url,
    keywords=keywords,
    license=license,
    classifiers=classifiers,
    include_package_data=include_package_data,
    #data_files=data_files,
    install_requires=install_requires,
    zip_safe=zip_safe,

    entry_points = {
        'nbconvert.exporters': [
            'xls = nb2xls:XLSExporter'
        ],
    }
)





def _get_version(version_info):
    dic = {'alpha': 'a',
           'beta': 'b',
           'candidate': 'rc',
           'final': ''}
    vi = version_info
    specifier = '' if vi[3] == 'final' else dic[vi[3]] + str(vi[4])
    version = '%s.%s.%s%s' % (vi[0], vi[1], vi[2], specifier)
    return version


# meta data - change alpha/dev to final for release
# also change in package.json

version_info = (0, 1, 0, 'final', 0)
__version__ = _get_version(version_info)

